#include "RtAudio.h"
#include "RtMidi.h"
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include <string.h>
#include <cstdlib>
#include <vector>

#define SAMPLE double
#define MY_PIE 3.14159265358979
#define MY_SRATE 48000

unsigned int maxDelayLength;
float testFreak = 200;
int activeNotes = 0;
int initialized = 0;
int addNoise = 0;
int numberOfVoices = 256; // starting polyphony

void usage( void ) {
	// Error function in case of incorrect command-line
	// argument specifications
	std::cerr << "\nuseage:	Pluck maxDelayLength\n";
	std::cerr << "	maxDelayLength: Sets the longest delay (in seconds) possible. Must be greater than 0.\n\n";
	
	exit( 1 );
}


struct strangThang {
	double * delayBuffer;
	double strangFreak; // delay frequency
	double attenFaktor;
	
	int writeHead;
	int readHead;
	int pluck; // if pluck = 1, add noise to the buffer
	int inUse;
};


// simple frequency to numSamples function
int freakToSamps(float hz)
{
	int samps = (int) MY_SRATE / hz;
	return samps;
}

// simple midi to frequency function
double mtof(int note)
{
	double frequency = 440 * pow(2.0,(note-69.0)/12.0);
	return frequency;
}







// The midi callback function
void getNotez( double deltatime, std::vector< unsigned char > *message, void *userData )
{
	
	std::vector<strangThang> * notes = (std::vector<strangThang> *) userData;
	
	unsigned int nBytes = message->size();
	
	// show midi bytes
	//for ( unsigned int i=0; i<nBytes; i++ )
		//std::cout << "Byte " << i << " = " << (int)message->at(i) << ", ";
	
	if (nBytes = 3) {
		if ((int)message->at(2) != 0 && (int)message->at(0) == 144) { // only look at note on messages with a non-zero velocity
			int midiPitch = (int)message->at(1);
			int midiVelocity = (int)message->at(2);
			double newFreak = mtof(midiPitch);
			int newNotePosition = 0; // newNotePosition is the address of the noteVector of our new note
			int more = 1;
			while ((*notes)[newNotePosition].inUse) {
				newNotePosition++;
				if (newNotePosition == numberOfVoices) {
					more = 0;
					}
			}
			if (more) {
				(*notes)[newNotePosition].strangFreak = newFreak;
				(*notes)[newNotePosition].readHead = 0;
				(*notes)[newNotePosition].writeHead =  freakToSamps(newFreak);
				(*notes)[newNotePosition].attenFaktor = midiVelocity / 127.1;
				(*notes)[newNotePosition].inUse = 1;
				(*notes)[newNotePosition].pluck = 1;
				std::cout << "Note Position: " << newNotePosition << std::endl;
				activeNotes++;
			}
		}
	}
	
	// show midi stamps
	//if ( nBytes > 0 )
	//	std::cout << "stamp = " << deltatime << std::endl;
}







// The audio callback function
int makeNoize( void *outputBuffer, void *inputBuffer, unsigned int nBufferFrames,
		double streamTime, RtAudioStreamStatus status, void *userData )
{	
	double * inBuff = (double *) inputBuffer;
	double * outBuff = (double *) outputBuffer;

	for ( int j=0 ; j<nBufferFrames ; j++ ) { // first clear the output buffer
		outBuff[j] = 0;
	}
	
	std::vector<strangThang> * notes = (std::vector<strangThang> *) userData;

	
	for ( int i = 0 ; i < (*notes).size(); i++ ) { // for all notes stored in noteArray
		if ((*notes)[i].inUse) {
			double ampAverage = 0;
			
			// Noise Generator
			if ((*notes)[i].pluck) {
				for ( int k = 0; k < maxDelayLength + 1; k++ ) { // for all samples in the delay buffer
					
					(*notes)[i].delayBuffer[k] = ((rand() / (double) RAND_MAX) * 2)  - 1;
				}
				(*notes)[i].pluck = 0;
				//std::cout << "plucked!" << std::endl;
			}			
			
			
			for ( int j=0 ; j<nBufferFrames ; j++ ) { // for all samples stored in the delay buffer
				(*notes)[i].delayBuffer[(*notes)[i].writeHead] = (.5 * ((*notes)[i].delayBuffer[(*notes)[i].readHead] + ((*notes)[i].delayBuffer[((*notes)[i].readHead - 1) % maxDelayLength])) * (*notes)[i].attenFaktor);
				(*notes)[i].writeHead = ((*notes)[i].writeHead + 1) % (maxDelayLength + 1); // advance writehead

				ampAverage += fabs((*notes)[i].delayBuffer[(*notes)[i].readHead]);
				//std::cout << ampAverage << std::endl;
				outBuff[j] += (*notes)[i].delayBuffer[(*notes)[i].readHead]; // store inst 1's delay buffer into audio output
				(*notes)[i].readHead = ((*notes)[i].readHead + 1) % (maxDelayLength + 1); // advance readhead
			}
			
			ampAverage = ampAverage / nBufferFrames;
			//std::cout << "Amp average: " << ampAverage << std::endl;
			if (ampAverage < pow(10, -3)) {
				(*notes)[i].inUse = 0;
				activeNotes--;
			}
		}
	}

	return 0;
}







// main function
int main( int argc, char ** argv )
{
	// Command Line parsing
	//---------------------
	if( argc != 2)
	{
		usage();
	}
	
	maxDelayLength = atoi(argv[1]);

	if (maxDelayLength <= 0) {
		std::cerr << "Max Delay Length must be positive:" << std::endl;
		usage();
	}
	
	
	// Initialize Note Vector for strangThang Objects
	//----------------------------------------------- 	
	
	std::vector <strangThang> noteVector (numberOfVoices); // how much polyphony
	
	for (int i = 0; i <numberOfVoices; i++) {
		noteVector[i].delayBuffer = new double[maxDelayLength];
		for (int j = 0; j < maxDelayLength; j++) {
			noteVector[i].delayBuffer[j] = 0;
		}
	}
	

	if (freakToSamps(testFreak) > maxDelayLength) {
		std::cerr << "Can't play notes that low, dummy! Set a longer Max Delay Length." << std::endl;
		exit( 0 );
	}
	
	noteVector[0].writeHead =  freakToSamps(testFreak);
	//std::cout << noteVector[0].readHead << " " << noteVector[0].writeHead << std::endl;
	
	
	// Set up RtAudio junk
	//--------------------
	RtAudio dac;
	if ( dac.getDeviceCount() < 1 ) {
		std::cout << "\nNo audio devices found!\n";
		exit( 0 );
	}
	
	RtAudio::StreamParameters Oparameters, Iparameters;
	Oparameters.deviceId = dac.getDefaultOutputDevice();
	Oparameters.nChannels = 1;
	Oparameters.firstChannel = 0;
	Iparameters.deviceId = 0;
	Iparameters.nChannels = 1;
	unsigned int sampleRate = 48000;
	unsigned int bufferFrames = 256; // 256 sample frames

	
	RtAudio::StreamOptions options;
	
	try {
		dac.openStream( &Oparameters, &Iparameters, RTAUDIO_FLOAT64,
					   sampleRate, &bufferFrames, &makeNoize, (void *) &noteVector /*(void *) noteVector*/ );
		dac.startStream();
	}
	catch ( RtError& e ) {
		e.printMessage();
		exit( 0 );
	}
	
	// Set up RtMidi Junk	
	//-------------------
	RtMidiIn *midiin = 0;
	// RtMidiIn constructor
	try {
		midiin = new RtMidiIn();
	}
	catch ( RtError &error ) {
		error.printMessage();
		exit( EXIT_FAILURE );
	}
	// Check available ports vs. specified.
	unsigned int nPorts = midiin->getPortCount();
	std::cout << "Number of ports: " << nPorts << std::endl;
	if ( nPorts > 0) {
		try {
			midiin->openPort( nPorts - 1 );
		}
		catch ( RtError &error ) {
			error.printMessage();
			goto cleanup;
		}
		midiin->setCallback( &getNotez , (void *) &noteVector /* (void *) noteArray */ );
		std::cerr << "Midi keyboard found, play away!\n" << std::endl;
	}
	else {
		std::cerr << "I can't seem to find any midi ports...\n" << std::endl;
		goto cleanup;
	}
	
	char input;
	while (!input) {
		std::cout << "Playing ... press <enter> to quit.\n";
		std::cin.get( input );
		//addNoise = 1;		
	}

	// Clean up
cleanup:
	try {
		// Stop the stream
		dac.stopStream();
	}
	catch (RtError& e) {
		e.printMessage();
	}
	
	if ( dac.isStreamOpen() ) dac.closeStream();
	delete midiin;

	return 0;
}
